<?php

	App::uses('VebraAppModel', 'Vebra.Model');
	App::uses('Branch', 'Vebra.Model');

	/**
	 * Description of Vebra
	 *
	 * @author Luke
	 */
	class Vebra extends VebraAppModel {

		public $useDbConfig = 'vebra';
		public $useTable = false;
		public $hasOne = array('Vebra.VebraToken');
		public $hasMany = array('Vebra.VebraLog');

		const FILE_TYPE_IMAGE = 0;
		const FILE_TYPE_FLOORPLAN = 2;
		const FILE_TYPE_BROCHURE = 7;
		const FILE_TYPE_EPC = 9;

		/**
		 * Look for an active token in the database and return it or return false.
		 * Note, this will only return the token - not the whole array.
		 * We don't cache the value because it's time sensitive, and storing the expiry time in three places isn't a good idea
		 * @return boolean|string
		 */
		public function getActiveToken() {
			$active_token = $this->VebraToken->findActive();
			return ($active_token !== false && !empty($active_token) ? $active_token['VebraToken']['token'] : false);
		}

		/**
		 * Get all branches from the API. VebraApi returns and array of branches, each with the following fields:
		 * - name
		 * - branchid
		 * - firmid
		 * - url
		 * - street
		 * - town
		 * - county
		 * - postcode
		 * - phone
		 * - email
		 */
		public function getBranches() {
			$this->_initialiseConnection(); // get a connection
			$branches = $this->getDataSource()->sendRequest('branch', $this->getActiveToken()); // send the request (returned in array format)
			if (is_array($branches['branch']) && !empty($branches['branch'])) { // if the response is a non-empty array
				$branches_with_details = array();
				foreach ($branches['branch'] as $branch) { // for each branch found
					preg_match('/\/([0-9]+)$/', $branch['url'], $matches); // get the branch ID from the end of the URL. Yeah, you'd think you could just use the branch id we get back, wouldn't you? 					$details = $this->_getBranchDetails($branch['url']); // get the branch details
					$details['BranchID'] = $branch['branchid'] = $matches[1]; // set both branch ID's (yeah, different cases, for laughs) to the one at the end fo the URL
					$branches_with_details[] = array_change_key_case(array_merge($branch, $details), CASE_LOWER); // merge the branch with the branch details (keys to lower case so we don't get dupes)
				}
				$this->_logRequest('get_all_branches');
				return $branches_with_details;
			}
			return false;
		}

		/**
		 * Takes a full URL (because the API returns the full URL for 'convenience') and gets
		 * a full set of details for a branch. Typically used when looping through branches
		 * @param string $url
		 * @return type
		 */
		protected function _getBranchDetails($url) {
			return $this->getDataSource()->sendRequest($url, $this->getActiveToken());
		}

		public function getProperties($branchId) {
			$this->_initialiseConnection();
			$properties = $this->getDataSource()->sendRequest('branch/' . $branchId . '/property', $this->getActiveToken());
			if (is_array($properties['property']) && !empty($properties['property'])) {
				$properties_with_details = array();
				$c = 0;
				foreach ($properties['property'] as $property) {
					// Limit properties imported when on the development environment (so it's quicker to test & debug)
					if (Configure::read('debug') == 2 && $c === 1) {
						break;
					}

					$branch = (new Branch)->find('first', array('conditions' => array('branchid' => $branchId)));
					$property['branch_id'] = $branch['Branch']['id'];

					$details = $this->_getPropertyDetails($property['url']);

					// Skip land plots
					if (preg_match('/^Land/', $details['type'][0])) {
						continue;
					}

					// Get the address details
					if (isset($details['address'])) {
						foreach ($details['address'] as $key => $value) {
							$property[$key] = $value;
						}
					}

					// Fix uploaded date, because they do it weird
					list($day, $month, $year) = explode('/', $details['uploaded']);
					$details['uploaded'] = implode('-', array($year, $month, $day));

					// Make sure no null values go into the database
					$details['bedrooms'] = (!isset($details['bedrooms']) || $details['bedrooms'] === null ? 0 : $details['bedrooms']);
					$details['receptions'] = (!isset($details['receptions']) || $details['receptions'] === null ? 0 : $details['receptions']);
					$details['bathrooms'] = (!isset($details['bathrooms']) || $details['bathrooms'] === null ? 0 : $details['bathrooms']);
					$details['garden'] = (!isset($details['garden']) || !$details['garden'] ? 0 : $details['garden']);
					$details['parking'] = (!isset($details['parking']) || !$details['parking'] ? 0 : $details['parking']);

					// Set the type
					$details['type'] = (isset($details['type'][0]) && !empty($details['type'][0]) ? $details['type'][0] : null);

					$property['id'] = $property['prop_id']; // use the Vebra prop_id as our primary key (to save useless auto-increment data)
					$properties_with_details[$c]['Property'] = array_change_key_case(array_merge($property, $details), CASE_LOWER);

					// Save the paragraphs (as associations in cake)
					if (!empty($details['paragraphs']['paragraph'])) {
						foreach ($details['paragraphs']['paragraph'] as $paragraph) {
							// This shouldn'y cause a conflict with any other array keys, given the data that we have at the moment, but if you ever lose data at the paragraph level, look into this
							// Merge paragraph attributes into the paragraph itself
							if (isset($paragraph['@attributes'])) {
								foreach ($paragraph['@attributes'] as $key => $value) {
									$paragraph[$key] = $value;
								}
								unset($paragraph['@attributes']);
							}

							// Merge the dimensions into the paragraph itself
							if (isset($paragraph['dimensions'])) {
								foreach ($paragraph['dimensions'] as $key => $value) {
									if (is_array($value) || empty($value)) { // if the value is an array, skip it
										continue;
									}
									$paragraph['dimensions_' . $key] = $value;
								}
								unset($paragraph['dimensions']);
							}
							if (isset($paragraph['id'])) {
								unset($paragraph['id']); // We don't want to mess with the primary key auto incrementing
							}
							$properties_with_details[$c]['Paragraph'][] = $paragraph; // we add directly to the merged array because we need to keep the casing for 'Paragraph' (model associations)
						}
					}

					if (!empty($details['files']['file'])) {
						foreach ($details['files']['file'] as $file) {
							switch ($file['@attributes']['type']) {
								case static::FILE_TYPE_IMAGE:
									$this->_importImage($file, $properties_with_details[$c]);
									break;

								case static::FILE_TYPE_FLOORPLAN:
									// Floor plans are also just images, but need to be put under the right header for the CMS management to work
									$this->_importImage($file, $properties_with_details[$c], 'Floorplan');
									break;

								case static::FILE_TYPE_BROCHURE:
									// This is a PDF document, so clearly can't be uploaded in the same way as images
									$this->_importDocument($file, $properties_with_details[$c], 'Brochure');
									break;

								case static::FILE_TYPE_EPC:
									// EPCs are also just images, and there is typically only one of them, though don't restrict it as such
									$this->_importImage($file, $properties_with_details[$c], 'Epc');
									break;
							}
						}
					}

					$c++;
				}

				$this->_logRequest('get_branch_properties', $branchId);
				return $properties_with_details;
			}
			return false;
		}

		public function getUpdatedProperties() {
			$this->_initialiseConnection();
			$response = $this->getDataSource()->sendRequest('property', $this->getActiveToken());
			return $response;
		}

		protected function _getPropertyDetails($url) {
			return $this->getDataSource()->sendRequest($url, $this->getActiveToken());
		}

		protected function _importImage($image, &$property, $attachment_type = 'Image') {
			preg_match('/([^\/]+\.([a-zA-Z]{3,4}))$/', $image['url'], $matches);
			$filename = $matches[1];
			$extension = $matches[2];
			switch ($extension) {
				case 'jpg' || 'jpeg':
					$type = 'image/jpeg';
					break;

				case 'png':
					$type = 'image/png';
					break;

				case 'gif':
					$type = 'image/gif';
					break;

				default:
					throw new Exception('Image type not supported');
					break;
			}

			// Store the temp image to upload shortly
			$tmp_name = rtrim(ini_get('upload_tmp_dir'), DS) . DS . $filename;
			copy($image['url'], $tmp_name);

			// Store the image using the regular upload method (no need to do anything special with it)
			$imageData = array(
				'model' => 'Property',
				'filename' => array(
					'name' => $filename,
					'type' => $type,
					'size' => filesize($tmp_name),
					'tmp_name' => $tmp_name,
					'error' => 0,
				),
				'attachment_type' => ($attachment_type !== 'Image' ? $attachment_type . 'Image' : $attachment_type),
				'type' => $type,
			);
			$property['Image'][] = $imageData; // add image to the images array, ready for saving
			return $property;
		}

		protected function _importDocument() {
			return true;
		}

		protected function _storeToken($token) {
			$this->VebraToken->create();
			$this->VebraToken->set('token', base64_encode($token));
			$this->VebraToken->save();
			$this->VebraToken->clear();
			return $this;
		}

		protected function _logRequest($action, $branchId = null, $propertyId = null, $paragraphId = null, $documentId = null) {
			$guzzleRequest = $this->getDataSource()->getGuzzleRequest();
			$url = $guzzleRequest->getUrl();
			$guzzleResponse = $this->getDataSource()->getGuzzleResponse();
			$guzzleException = $this->getDataSource()->getGuzzleException();

			$vebraLog = array();
			$vebraLog['VebraLog']['action'] = $action;
			$vebraLog['VebraLog']['branch_id'] = $branchId;
			$vebraLog['VebraLog']['property_id'] = $propertyId;
			$vebraLog['VebraLog']['paragraph_id'] = $paragraphId;
			$vebraLog['VebraLog']['document_id'] = $documentId;
			$vebraLog['VebraLog']['url'] = $url;
			$vebraLog['VebraLog']['request_header'] = json_encode($guzzleRequest->getHeaders());
			$vebraLog['VebraLog']['response_header'] = json_encode(($guzzleResponse !== null ? $guzzleResponse->getHeaders() : $guzzleException->getRequest()->getResponse()->getHeaders()));
			$vebraLog['VebraLog']['response_body'] = json_encode(($guzzleResponse !== null ? $guzzleResponse->getBody() : $guzzleException->getRequest()->getResponse()->getBody()));
			$vebraLog['VebraLog']['response_code'] = json_encode(($guzzleResponse !== null ? $guzzleResponse->getStatusCode() : $guzzleException->getRequest()->getResponse()->getStatusCode()));
			$this->VebraLog->create();
			$this->VebraLog->save($vebraLog);

			return $this;
		}

		/**
		 * Get a new active token if none exist currently.
		 * This can be pasted atop every method that needs authentication, and it'll take care of getting you a new token when needed.
		 * It works because we don't store the token in the class - we get it from the database each time. This ensures that the API doesn't fall over part-way through
		 * @return Vebra
		 */
		protected function _initialiseConnection() {
			if ($this->getActiveToken() === false) {
				$token = $this->getDataSource()->getNewToken();
				$this->_storeToken($token);
			}
			return $this;
		}

	}
